
# -*- coding: utf-8 -*-
"""
手机屏幕截图的代码
"""
import os
import subprocess
import sys
from io import StringIO

from PIL import Image

from .auto_adb import AutoADB

adb = AutoADB()
# SCREENSHOT_WAY 是截图方法，经过 check_screenshot 后，会自动递减，不需手动修改
SCREENSHOT_WAY = 3


def pull_screenshot():
    """
    获取屏幕截图，目前有 0 1 2 3 四种方法，未来添加新的平台监测方法时，
    可根据效率及适用性由高到低排序
    """
    global SCREENSHOT_WAY
    if 1 <= SCREENSHOT_WAY <= 3:
        process = subprocess.Popen(
            adb.adb_path + ' shell screencap -p',
            shell=True, stdout=subprocess.PIPE)
        binary_screenshot = process.stdout.read()
        print(binary_screenshot)
        if SCREENSHOT_WAY == 2:
            binary_screenshot = binary_screenshot.replace(b'\r\n', b'\n')
        elif SCREENSHOT_WAY == 1:
            binary_screenshot = binary_screenshot.replace(b'\r\r\n', b'\n')
        return Image.open(StringIO(binary_screenshot))
    elif SCREENSHOT_WAY == 0:
        adb.cate_hot_list('shell screencap -p /sdcard/autojump.png')
        adb.cate_hot_list('pull /sdcard/autojump.png .')
        return Image.open('./autojump.png')


def check_screenshot():
    """
    检查获取截图的方式
    """
    global SCREENSHOT_WAY
    if os.path.isfile('autojump.png'):
        try:
            os.remove('autojump.png')
        except Exception:
            pass
    if SCREENSHOT_WAY < 0:
        print('暂不支持当前设备')
        sys.exit()
    try:
        im = pull_screenshot()
        im.load()
        im.close()
        print('采用方式 {} 获取截图'.format(SCREENSHOT_WAY))
    except Exception:
        SCREENSHOT_WAY -= 1
        check_screenshot()


# check_screenshot()
# pull_screenshot()



# -*- coding: utf-8 -*-
"""
Simple algorithm to solve nonograms using left and right overlaps

See details:
http://www.lancaster.ac.uk/~simpsons/nonogram/ls-fast
"""

from notetool.tool.log import logger
from six.moves import range

from ..core.common import BOX, SPACE, UNKNOWN, NonogramError
from .base import BaseLineSolver

_SYMBOL_MAP = {
    UNKNOWN: ' ',
    SPACE: '-',
    BOX: '#',
}


class FastSolver(BaseLineSolver):
    """
    Nonogram line solver that uses left and right overlap algorithm.
    The algorithm gets most of the solution, but sometimes not complete one.
    """

    @classmethod
    def push_left(cls, line, clue):
        """
        Move all the blocks to the left.
        Raise NonogramError if inconsistency detected.
        """

        line_size = len(line)
        clue_size = len(clue)
        res = [0] * clue_size
        solid = [-1] * clue_size

        # initial state
        current_block = 0

        if clue_size > 0:
            res[current_block] = 0

        logger.info('Pushing clue: %s', ', '.join(map(str, clue)))
        logger.info('Pushing line: >%s<', ''.join(
            _SYMBOL_MAP.get(cell, '?') for cell in line))

        while current_block < clue_size:
            # find first/next non-dot:
            # stop if current_block won't fit into remainder of line

            pos = res[current_block]
            block_size = clue[current_block]

            logger.debug('     start %d >%s %d',
                         current_block, pos, block_size)

            while pos < line_size - block_size:
                cell = line[pos]
                if cell != SPACE:
                    break

                pos += 1

            res[current_block] = pos
            logger.debug('     end %d >%s %d', current_block, pos, block_size)

            # no room left
            if (pos + block_size > line_size) or (line[pos] == SPACE):
                raise NonogramError(
                    'No room left: cannot fit %d-th block' % current_block)

            # assume current position doesn't cover a solid
            solid[current_block] = -1

            # check if the current_block fits in before the next dot;
            # monitor for passing over a solid
            i = 0
            while i < block_size:
                cell = line[pos + i]
                if cell == SPACE:
                    break

                if (solid[current_block] < 0) and (cell == BOX):
                    logger.debug('     solid %d >%s#',
                                 current_block, (pos + i))
                    solid[current_block] = i
                i += 1

            # if a dot was encountered...
            if i < block_size:
                logger.debug('     dot %d >%s-', current_block, (pos + i))

                # if a solid is covered, get an earlier current_block to fit
                if solid[current_block] >= 0:
                    # find an earlier current_block that can be moved to cover the solid
                    # covered by the next current_block without uncovering its own
                    while True:
                        if current_block == 0:
                            raise NonogramError(
                                'All previous blocks cover solids')
                        current_block -= 1
                        if (solid[current_block] < 0) or (
                                res[current_block + 1] + solid[current_block + 1] -
                                clue[current_block] + 1 <=
                                res[current_block] + solid[current_block]):
                            break

                    # set the current_block near to the position of the next current_block,
                    # so that it just overlaps the solid, and try again
                    next_pos = res[current_block + 1]
                    next_solid = solid[current_block + 1]
                    res[current_block] = next_pos + \
                        next_solid - clue[current_block] + 1
                    continue

                # otherwise, simply move to the dot, and try again
                res[current_block] += i
                continue

            # now check to see if the end of the current current_block touches an
            # existing solid - if so, the current_block must be shuffled further to
            # ensure it overlaps the solid, but no solid may emerge from the
            # other end
            pos = res[current_block]

            if pos + block_size < line_size:
                end_block_cell = line[pos + block_size]

                if (end_block_cell == BOX) and (solid[current_block] < 0):
                    solid[current_block] = block_size

                while pos + block_size < line_size:
                    cell = line[pos]
                    if cell == BOX:
                        break
                    end_block_cell = line[pos + block_size]
                    if end_block_cell != BOX:
                        break

                    pos += 1
                    solid[current_block] -= 1

            res[current_block] = pos

            logger.debug('     shuffle %s >%s %d',
                         current_block, pos, block_size)

            # if there's still a solid immediately after the current_block, there's
            # an error, so find an earlier current_block to move
            if pos + block_size < line_size:
                end_block_cell = line[pos + block_size]
                if end_block_cell == BOX:
                    logger.debug('     stretched %d >%s#',
                                 current_block, (pos + i))

                    # find an earlier current_block that isn't covering a solid
                    error_block = current_block
                    while True:
                        if current_block == 0:
                            raise NonogramError(
                                'The %d-th block cannot be stretched' % error_block)
                        current_block -= 1
                        if (solid[current_block] < 0) or (
                                res[current_block + 1] + solid[current_block + 1] -
                                clue[current_block] + 1 <=
                                res[current_block] + solid[current_block]):
                            break

                    # set the current_block near to the position of the next current_block,
                    # so that it just overlaps the solid, and try again
                    next_pos = res[current_block + 1]
                    next_solid = solid[current_block + 1]
                    res[current_block] = next_pos + \
                        next_solid - clue[current_block] + 1
                    continue

            # the current_block is in place, so try the next
            pos = res[current_block] + 1 + clue[current_block]
            if current_block + 1 < clue_size:
                current_block += 1
                res[current_block] = pos
            else:
                # no more blocks, so just check for any remaining solids
                while pos < line_size:
                    cell = line[pos]
                    if cell == BOX:
                        break

                    pos += 1

                # if a solid was found...
                if pos < line_size:
                    logger.debug('     trailing >%s#', pos)

                    # move the current_block so it covers it, but check if solid
                    # becomes uncovered
                    if (solid[current_block] >= 0) and (
                            pos - block_size + 1 > res[current_block] + solid[current_block]):
                        #  a solid was uncovered, so look for an earlier current_block

                        # find an earlier current_block that isn't covering a solid
                        while True:
                            if current_block == 0:
                                raise NonogramError(
                                    'All previous blocks cover solids')
                            current_block -= 1
                            if (solid[current_block] < 0) or (
                                    res[current_block + 1] +
                                    solid[current_block + 1] - clue[current_block] + 1 <=
                                    res[current_block] + solid[current_block]):
                                break

                        # set the current_block near to the position of the next current_block,
                        # so that it just overlaps the solid, and try again
                        next_pos = res[current_block + 1]
                        next_solid = solid[current_block + 1]
                        res[current_block] = next_pos + \
                            next_solid - clue[current_block] + 1

                        continue
                    # if (solid[current_block] >= 0) ... }

                    # solid[current_block] -= pos - block_size + 1 - res[current_block]
                    res[current_block] = pos - block_size + 1
                    continue
                # if pos < line_size ... }
                # no solid found
                current_block += 1
            # if current_block + 1 < clue_size: else ... }
        # while current_block < clue_size ... }
        # all okay
        return res

    @classmethod
    def push_right(cls, line, clue):
        """Move all the blocks to the right"""
        clue = clue[::-1]
        line = line[::-1]

        return list(reversed(cls.push_left(line, clue)))

    def _solve(self):
        """
        Solve the given line with given clue (description)
        using left and right overlap algorithm
        """

        line = self.line
        clue = self.description

        line_size = len(line)
        clue_size = len(clue)

        logger.debug('Line range = %s to %s', 0, line_size - 1)

        if (clue_size == 1) and clue[0] == 0:
            clue_size = 0

        left_positions = self.push_left(line, clue)

        left_desc = []
        for block in range(clue_size):
            left_desc.append('(%s + %s)' %
                             (left_positions[block], clue[block]))
        logger.info('Left: %s', ' '.join(left_desc))
        left_desc = []
        for block in range(clue_size):
            size_now = sum(map(len, left_desc))
            block = (
                '-' * (left_positions[block] - size_now)) + ('#' * clue[block])
            left_desc.append(block)

        desc_size = len(''.join(left_desc))
        left_desc.append('-' * (line_size - desc_size))
        logger.info('Left: >%s<', ''.join(left_desc))

        middle_desc = [_SYMBOL_MAP.get(cell, '?') for cell in line]
        logger.info('Middle: >%s<', ''.join(middle_desc))

        logger.info('Line range = %d to %d', 0, line_size - 1)

        right_positions = self.push_right(line, clue)

        right_desc = []
        for block in range(clue_size):
            right_desc.append('(%s + %s)' %
                              (right_positions[block], clue[block]))
        logger.info('Right: %s', ' '.join(right_desc))
        right_desc = []
        for block in range(clue_size):
            size_now = sum(map(len, right_desc))
            dots = '-' * \
                (line_size - right_positions[block] - clue[block] - size_now)
            solids = '#' * (clue[block])
            block = dots + solids
            right_desc.append(block)

        desc_size = len(''.join(right_desc))
        right_desc.append('-' * (line_size - desc_size))
        logger.info('Right: >%s<', ''.join(right_desc))

        work = list(line)

        j = 0
        for i in range(clue_size):
            right_positions[i] = line_size - right_positions[i] - clue[i]

            k = left_positions[i]
            work[j: k] = [SPACE] * (k - j)

            j = right_positions[i]
            k = left_positions[i] + clue[i]

            work[j: k] = [BOX] * (k - j)

            j = right_positions[i] + clue[i]

        k = line_size
        work[j: k] = [SPACE] * (k - j)

        return work



# -*- coding: utf-8 -*-


import re
import string
from collections import OrderedDict, namedtuple

from notetool.tool.log import logger
from six import integer_types, itervalues, string_types

from ..utils.iter import expand_generator


class Color(object):
    """
    Aggregate a single color different representations

    name: used for referring in clues (e.g. examples/uk.txt)
    rgb: color code to use in SVG (can be a common color name like 'green')
    symbol: ASCII symbol to show the color in terminal
    id: integer, used in solvers
    """

    def __init__(self, name, rgb, symbol=None, id_=None):
        self.name = name
        self.rgb = rgb
        self.symbol = symbol
        self.id_ = id_

    def __repr__(self):
        return '{}({!r}, {!r}, symbol={!r}, id_={!r})'.format(
            self.__class__.__name__,
            self.name, self.rgb, self.symbol, self.id_)

    @classmethod
    def white(cls):
        """Predefined white color (space) representation"""
        return cls('white', 'fff', ' ', 1)

    @classmethod
    def black(cls):
        """Predefined black color (box) representation"""
        return cls('black', '000', 'X', 2)

    RGB_TRIPLET_RE = re.compile(r'([0-9]+),[ \t]*([0-9]+),[ \t]*([0-9]+)')

    @property
    def svg_name(self):
        """Generate an SVG-readable description for a color"""

        rgb = self.rgb
        if isinstance(rgb, (list, tuple)) and len(rgb) == 3:
            return 'rgb({})'.format(','.join(map(str, rgb)))

        if len(rgb) in (3, 6) and all(letter in string.hexdigits for letter in rgb):
            return '#' + rgb

        if isinstance(rgb, string_types) and self.RGB_TRIPLET_RE.match(rgb):
            return 'rgb({})'.format(rgb)

        return rgb


class ColorMap(OrderedDict):
    """
    Store the collection of colors.
    Encapsulate creation, search and iteration among them.
    """

    def __init__(self):
        """Prevent creating from other iterables"""

        super(ColorMap, self).__init__()
        self.by_id = dict()

        for color in [Color.white(), Color.black()]:
            self.push_color(color)

        # only black and white are added ny now
        self.black_and_white = True

    def iter_colors(self):
        """Iterate over stored colors"""
        return itervalues(self)

    _MIN_ID = 1 << 2
    _SYMBOLS = string.punctuation + string.ascii_uppercase + string.ascii_lowercase

    def _new_symbol(self):
        registered_symbols = set(color.symbol for color in self.iter_colors())

        for symbol in self._SYMBOLS:
            if symbol not in registered_symbols:
                return symbol

        return None

    def push_color(self, color):
        """
        Add a color to the map
        :type color: Color
        """
        self[color.name] = color
        # for every added new color, set the map as colored
        self.black_and_white = False

    def _new_id(self):
        if self:
            max_id = max(color.id_ for color in self.iter_colors())
            # every color is a '100...' in binary
            # that was made to allow bitwise OR to signify multiple colors
            return max_id << 1
        return self._MIN_ID

    @classmethod
    def _check_id(cls, id_):
        id_ = int(id_)
        if id_ < cls._MIN_ID:
            raise ValueError('Bad id_ value: {!r}'.format(id_))

    def make_color(self, name, rgb, symbol=None, id_=None):
        """
        Make a new color, optionally filling up id and symbol
        """

        if name in self:
            color = self[name]
            logger.info('Color %r already found: %r', name, color)

            color.rgb = rgb
            if symbol is not None:
                color.symbol = symbol
            if id_ is not None:
                self._check_id(id_)
                color.id_ = id_

            return color

        if symbol is None:
            symbol = self._new_symbol()
            if symbol is None:
                raise ValueError(
                    'No available symbols left for color {!r}'.format(name))

        if id_ is None:
            id_ = self._new_id()

        self._check_id(id_)

        new_color = Color(name, rgb, symbol=symbol, id_=id_)
        self.push_color(new_color)
        return new_color

    def find_by_id(self, id_):
        """
        Find color by given color id
        """
        return self.by_id.get(id_)

    def find_by_name(self, name):
        """
        Find color by given color name
        """
        return self.get(name)

    def __setitem__(self, key, value):
        super(ColorMap, self).__setitem__(key, value)
        try:
            self.by_id[value.id_] = value
        except AttributeError:
            raise ValueError(
                'Only colors can be set as values in {!r}'.format(self.__class__))


class ColorBlock(namedtuple('ColorBlock', 'size color')):
    """Represent one block of colored description"""


_COLOR_DESCRIPTION_RE = re.compile('([0-9]+)(.+)')


@expand_generator(type_=tuple)
def normalize_description_colored(row, color_map):
    """Normalize a colored nonogram description"""

    row = normalize_description(row, color=True)

    black_color = Color.black().name

    res = []
    for block in row:
        item = None
        if isinstance(block, integer_types):
            item = (block, black_color)
        elif isinstance(block, string_types):
            match = _COLOR_DESCRIPTION_RE.match(block)
            if match:
                item = (int(match.group(1)), match.group(2))
            else:
                item = (int(block), black_color)
        elif isinstance(block, (tuple, list)) and len(block) == 2:
            size = block[0]
            if size != BlottedBlock:
                size = int(size)

            item = (size, block[1])

        if item is None:
            raise ValueError('Bad description block: {}'.format(block))
        else:
            res.append(item)

    for size, color_name in res:
        if color_name in color_map.by_id:
            id_ = color_name
        else:
            id_ = color_map[color_name].id_
        yield ColorBlock(size, id_)



# -*- coding: utf-8 -*-
"""
Defines various renderers for the game of nonogram
"""

from abc import ABC
from sys import stdout

from notetool.tool.log import logger
from six import integer_types, itervalues, text_type

from ..utils.iter import max_safe, pad
from ..utils.other import two_powers
from .common import BOX, SPACE, UNKNOWN, BlottedBlock, is_list_like


class Cell(object):
    """Represent basic rendered cell"""

    DEFAULT_ICON = ' '

    def __init__(self, icon=None):
        self.icon = icon or self.DEFAULT_ICON

    def ascii_icon(self):
        """How the cell can be printed as a text"""
        return self.DEFAULT_ICON

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)


class ThumbnailCell(Cell):
    """
    Represent upper-left cell
    (where the thumbnail of the puzzle usually drawn).
    """
    DEFAULT_ICON = '#'


class ClueCell(Cell):
    """
    Represent cell that is part of description (clue).
    They are usually drawn on the top and on the left.
    """

    BLOTTED_SYMBOL = '?'

    def __init__(self, value):
        super(ClueCell, self).__init__()
        if is_list_like(value):
            self.value, self.color = value
        else:
            self.value, self.color = value, None

    def ascii_icon(self):
        """
        Gets a symbolic representation of a cell given its state
        and predefined table `icons`
        """
        if isinstance(self.value, integer_types):
            return text_type(self.value)

        if self.value == BlottedBlock:
            return self.BLOTTED_SYMBOL

        return self.DEFAULT_ICON

    def __repr__(self):
        return '{}(({}, {}))'.format(
            self.__class__.__name__,
            self.value, self.color)


class GridCell(Cell):
    """Represent the main area cell"""

    def __init__(self, value, renderer, colored=False):
        super(GridCell, self).__init__()

        self.renderer = renderer
        self.colored = colored
        if self.colored:
            self.value = tuple(two_powers(value))
        else:
            self.value = value

    def ascii_icon(self):
        value = self.value
        icons = self.renderer.icons

        if not self.colored:
            return icons[self.value]

        if len(value) == 1:
            value = value[0]
        else:
            # multiple colors
            value = UNKNOWN

        symbol = self.renderer.board.symbol_for_color_id(value)
        if symbol is not None:
            return symbol

        return icons.get(value, self.DEFAULT_ICON)

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__, self.value)


class _DummyBoard(object):
    """
    Stub for renderer initialization
    when it created before the corresponding board
    """
    rows_descriptions = columns_descriptions = ()
    width = height = 0


class Renderer(object):
    """Defines the abstract renderer for a nonogram board"""

    def __init__(self, board=None):
        self.cells = None
        self.board = None
        self.board_init(board)

    def board_init(self, board=None):
        """Initialize renderer's properties dependent on board it draws"""
        if board:
            logger.info('Init %r renderer with board %r',
                        self.__class__.__name__, board)
        else:
            if self.board:
                return  # already initialized, do nothing
            board = _DummyBoard()
        self.board = board

    @property
    def full_height(self):
        """The full visual height of a board"""
        return self.header_height + self.board.height

    @property
    def full_width(self):
        """The full visual width of a board"""
        return self.side_width + self.board.width

    @property
    def header_height(self):
        """The size of the header block with columns descriptions"""
        return max_safe(map(len, self.board.columns_descriptions), default=0)

    @property
    def side_width(self):
        """The width of the side block with rows descriptions"""
        return max_safe(map(len, self.board.rows_descriptions), default=0)

    def render(self):
        """Actually print out the board"""
        raise NotImplementedError()

    def draw(self, cells=None):
        """Calculate all the cells and draw an image of the board"""
        self.draw_header()
        self.draw_side()
        self.draw_grid(cells=cells)
        self.render()

    def draw_header(self):
        """
        Changes the internal state to be able to draw columns descriptions
        """
        raise NotImplementedError()

    def draw_side(self):
        """
        Changes the internal state to be able to draw rows descriptions
        """
        raise NotImplementedError()

    def draw_grid(self, cells=None):
        """
        Changes the internal state to be able to draw a main grid
        """
        raise NotImplementedError()

    @property
    def is_colored(self):
        """Whether the linked board is colored board"""
        return self.board.is_colored


class StreamRenderer(Renderer, ABC):
    """
    Simplify textual rendering of a board to a stream (stdout by default)
    """

    DEFAULT_ICONS = {
        UNKNOWN: '_',
        BOX: 'X',
        SPACE: '.',
    }

    def __init__(self, board=None, stream=stdout, icons=None):
        self.stream = stream
        if icons is None:
            icons = dict(self.DEFAULT_ICONS)
        self.icons = icons
        super(StreamRenderer, self).__init__(board)

    def _print(self, *args):
        return print(*args, file=self.stream)


class BaseAsciiRenderer(StreamRenderer):
    """
    Renders a board as a simple text table (without grid)
    """

    __rend_name__ = 'text'

    def board_init(self, board=None):
        super(BaseAsciiRenderer, self).board_init(board)
        logger.info('init cells: %sx%s', self.full_width, self.full_width)

        self.cells = [[Cell()] * self.full_width
                      for _ in range(self.full_height)]

    def cell_icon(self, cell):
        """
        Get a symbolic representation of a cell given its state
        and predefined table `icons`
        """
        return cell.ascii_icon()

    def render(self):
        for row in self.cells:
            res = []
            for index, cell in enumerate(row):
                ico = self.cell_icon(cell)

                # do not pad the last symbol in a line
                if len(ico) == 1:
                    if index < len(row) - 1:
                        ico += ' '

                res.append(ico)

            self._print(''.join(res))

    def draw_header(self):
        for i in range(self.header_height):
            for j in range(self.side_width):
                self.cells[i][j] = ThumbnailCell()

        for j, col in enumerate(self.board.columns_descriptions):
            rend_j = j + self.side_width
            if not col:
                col = [0]

            rend_column = [ClueCell(val) for val in col]
            rend_column = pad(rend_column, self.header_height, Cell())

            # self.cells[:self.header_height, rend_j] = rend_column
            for i, cell in enumerate(rend_column):
                self.cells[i][rend_j] = cell

    def draw_side(self):
        for i, row in enumerate(self.board.rows_descriptions):
            rend_i = i + self.header_height
            # row = list(row)
            if not row:
                row = [0]

            rend_row = [ClueCell(val) for val in row]
            rend_row = pad(rend_row, self.side_width, Cell())
            self.cells[rend_i][:self.side_width] = rend_row

    def draw_grid(self, cells=None):
        if cells is None:
            cells = self.board.cells

        is_colored = self.is_colored

        for i, row in enumerate(cells):
            rend_i = i + self.header_height
            for j, val in enumerate(row):
                rend_j = j + self.side_width
                self.cells[rend_i][rend_j] = GridCell(
                    val, self, colored=is_colored)


def _register_renderers():
    res = dict()
    for obj in itervalues(globals()):
        if isinstance(obj, type):
            if issubclass(obj, StreamRenderer) and hasattr(obj, '__rend_name__'):
                res[obj.__rend_name__] = obj
    return res


RENDERERS = _register_renderers()



# -*- coding: utf-8 -*-
"""Define nonogram solving operations"""


import logging

from notegame.games.nonogram.core.common import (normalize_description,
                                                 normalize_row)

from notegame.games.nonogram.core.common import (normalize_description,                                                normalize_row)

from . import bgu, efficient, machine, simpson

# TODO: choose the method for each registered solver
SOLVERS = {
    'partial_match': machine.PartialMatchSolver,
    'reverse_tracking': machine.ReverseTrackingSolver,

    'simpson': simpson.FastSolver,

    'bgu': bgu.BguSolver,
    'blot': bgu.BguBlottedSolver,

    'efficient': efficient.EfficientSolver,
}


def solve_line(desc, line, method='reverse_tracking', normalized=False):
    """
    Utility for row solving that can be used in multiprocessing map
    """
    # method = kwargs.pop('method', 'reverse_tracking')
    #
    # if len(args) == 1:
    #     # mp's map supports only one iterable, so this weird syntax
    #     args = args[0]
    #
    # desc, line = args

    if not normalized:
        desc = normalize_description(desc)
        # desc = tuple(desc)
        line = normalize_row(line)

    try:
        solver = SOLVERS[method]
    except KeyError:
        raise KeyError("Cannot find solver '%s'" % method)

    return solver.solve(desc, line)



# -*- coding: utf-8 -*-


from notetool.tool.log import logger
from six import iteritems, text_type


class StateMachineError(ValueError):
    """
    Represents an error occurred when trying
    to make bad transition with a FSM
    """
    BAD_TRANSITION = 1
    BAD_ACTION = 2

    def __init__(self, *args, **kwargs):
        self.code = kwargs.pop('code', None)
        super(StateMachineError, self).__init__(*args)


class FiniteStateMachine(object):
    """
    Represents a simple deterministic automaton that can be
    only in a single state in every single moment

    https://en.wikipedia.org/wiki/Finite-state_machine
    """

    __slots__ = ['initial_state', '_state', 'state_map', 'final_state']

    def __init__(self, initial_state, state_map, final=None):
        self.initial_state = initial_state
        self._state = initial_state
        self.state_map = dict(state_map)
        self.final_state = final

        # the assertion never failed but took too long, so just switch it off
        # assert self.current_state in self.states

    def transition(self, *actions):
        """
        Change the state of a machine by consequently applying
        one or more `actions`
        """
        state = None
        for action in actions:
            state = self.transition_one(action)

        return state

    def transition_one(self, action):
        """
        Change the state of a machine according to the
        `self.state_map` by applying an `action`
        """
        logger.debug('Current state: %r', self.current_state)
        logger.debug('Action: %r', action)

        if action not in self.actions:
            raise StateMachineError("Action '{}' not available".format(
                action), code=StateMachineError.BAD_ACTION)

        new_state = self.reaction(action)
        if new_state is None:
            raise StateMachineError("Cannot do '{}' from the state '{}'".format(
                action, self.current_state), code=StateMachineError.BAD_TRANSITION)
        else:
            self._state = new_state
            logger.debug('New state: %r', self.current_state)
            return self.current_state

    def reaction(self, action, current_state=None):
        """
        Returns the state in which the machine would be
        if the `action` will apply to it
        (without actual changing of the state).

        If the `current_state` is given, pretend that
        the machine is in this state before action.
        """
        if current_state is None:
            current_state = self.current_state

        return self.state_map.get((current_state, action))

    @property
    def current_state(self):
        """
        The current state of a machine
        """
        return self._state

    @property
    def states(self):
        """
        All the possible states of a machine
        """
        return tuple(set(state for state, action in self.state_map))

    @property
    def actions(self):
        """
        All the possible actions that can be applied to a machine
        """
        return tuple(set(action for state, action in self.state_map))

    def __str__(self):
        res = [
            '{}({});'.format(self.__class__.__name__, self.current_state),
            'All states: [{}];'.format(
                ', '.join(sorted(map(text_type, self.states)))),
            'All actions: [{}];'.format(
                ', '.join(sorted(map(text_type, self.actions)))),
            'States map:',
        ]
        res.extend([
            '{}, {} -> {}'.format(state, action, new_state)
            for (state, action), new_state in sorted(iteritems(self.state_map))])

        if self.final_state is not None:
            res.append('Final state: {}.'.format(self.final_state))
        return '\n'.join(res)

    def match(self, word):
        """
        Verify if the machine can accept the `word`
        i.e. reach the `self.final_state` by applying
        the words' letters one by one
        """
        if self.current_state != self.initial_state:
            raise RuntimeError('Only run {!r} when in initial state {!r}'.format(
                self.match.__name__, self.initial_state))

        if self.final_state is None:
            raise RuntimeError('Cannot match: no final state defined')

        for letter in word:
            logger.debug('Match letter %r of word %r', letter, word)
            try:
                prev = self.current_state
                self.transition(letter)
                logger.info('Transition from %r to %r with action %r',
                            prev, self.current_state, letter)
            except StateMachineError:
                logger.info('Cannot do action %r in the state %r',
                            letter, self.current_state)
                return False

        return self.current_state == self.final_state

    def __copy__(self):
        new_one = type(self)(self.initial_state, self.state_map, final=self.final_state)
        return new_one



# -*- coding: utf-8 -*-
"""Define nonogram solver that solves line-by-line"""

import time

from notetool.tool.log import logger
from six.moves import range, zip

from ..solver import solve_line
from ..utils.priority_dict import PriorityDict
from .common import BOX, SPACE, UNKNOWN, is_color_cell


def _is_pixel_updated(old, new):
    if old == new:
        return False

    if is_color_cell(old):
        # old value contains more '1'-s in the binary representation
        assert old > new
    else:
        assert old == UNKNOWN
        assert new in (BOX, SPACE)

    return True


def solve_row(board, index, is_column, method):
    """
    Solve a line with the solving `method`.
    If the line gets partially solved,
    put the crossed lines into queue.

    Return the list of new jobs that should be solved next (one for each solved cell).
    """

    # start = time.time()

    if is_column:
        row_desc = board.columns_descriptions[index]
        row = tuple(board.get_column(index))
        # desc = 'column'
    else:
        row_desc = board.rows_descriptions[index]
        row = tuple(board.get_row(index))
        # desc = 'row'

    # pre_solution_rate = board.line_solution_rate(row)

    # if board.is_line_solved(row):
    #     # do not check solved lines in trusted mode
    #     if contradiction_mode:
    #         assert_match(row_desc, row)
    #     return 0, ()

    # LOG.debug('Solving %s %s: %s. Partial: %s', index,
    #           'column' if is_column else 'row', row_desc, row)

    updated = solve_line(row_desc, row, method=method, normalized=True)

    new_jobs = []

    # if board.line_solution_rate(updated) > pre_solution_rate:
    if row != updated:
        # LOG.debug('Queue: %s', jobs_queue)
        # LOG.debug(row)
        # LOG.debug(updated)
        for i, (pre, post) in enumerate(zip(row, updated)):
            if _is_pixel_updated(pre, post):
                new_jobs.append((not is_column, i))
        # LOG.debug('Queue: %s', jobs_queue)
        # LOG.debug('New info on %s %s: %s', desc, index, [job_index for _, job_index in new_jobs])

        if is_column:
            board.set_column(index, updated)
        else:
            board.set_row(index, updated)

    # LOG.debug('%ss solution: %.6f sec', desc, time.time() - start)
    return new_jobs


def solve(board,
          row_indexes=None, column_indexes=None,
          contradiction_mode=False, methods=None):
    """
    Solve the nonogram to the most using two methods (by default):
    - firstly with simple right-left overlap algorithm
    - then with FSM and reverse tracking

    All methods use priority queue to store the lines needed to solve.

    Return the total number of solved cells.
    """

    if methods is None:
        if board.is_colored:
            if board.has_blots:
                methods = ('blot_color',)
            else:
                methods = ('bgu_color',)
        else:
            if board.has_blots:
                methods = ('blot',)
            else:
                # methods = ('simpson', 'reverse_tracking')
                methods = ('bgu',)

    if not isinstance(methods, (tuple, list)):
        methods = [methods]

    total_cells_solved = 0
    for method in methods:
        cells_solved, jobs = _solve_with_method(
            board, method,
            row_indexes=row_indexes,
            column_indexes=column_indexes,
            contradiction_mode=contradiction_mode)

        total_cells_solved += cells_solved
        row_indexes = [index for is_column, index in jobs if not is_column]
        column_indexes = [index for is_column, index in jobs if is_column]

    return total_cells_solved


def _solve_with_method(
        board, method,
        row_indexes=None, column_indexes=None,
        contradiction_mode=False):
    """Solve the nonogram to the most using given method"""

    # `is_solved_full` is cost, so minimize calls to it.
    # Do not call if only a handful of lines has to be solved
    if row_indexes is None or column_indexes is None or \
            len(row_indexes) > 2 or len(column_indexes) > 2:

        # do not shortcut in contradiction_mode
        if not contradiction_mode and board.is_solved_full:
            return 0, ()

    has_blots = board.has_blots

    start = time.time()
    lines_solved = 0

    # every job is a tuple (is_column, index)
    #
    # Why `is_column`, not `is_row`?
    # To assign more priority to the rows:
    # when adding row, `is_column = False = 0`
    # when adding column, `is_column = True = 1`
    # heap always pops the lowest item, so the rows will go first

    logger.debug('Solving %s rows and %s columns with %r method' %
                 (row_indexes, column_indexes, method))

    line_jobs = PriorityDict()
    all_jobs = set()

    def _add_job(job, _priority):
        line_jobs[job] = _priority
        all_jobs.add(job)

    if row_indexes is None:
        row_indexes = range(board.height)
    for row_index in row_indexes:
        # the more this line solved
        # priority = 1 - board.row_solution_rate(row_index)

        # the closer to edge
        # priority = 1 - abs(2.0 * row_index / board.height - 1)

        # the more 'dense' this line
        # priority = 1 - board.densities[False][row_index]

        new_job = (False, row_index)

        priority = 0
        if has_blots:
            # the more attempts the less priority
            priority = board.attempts_to_try(*new_job)

        _add_job(new_job, priority)

    if column_indexes is None:
        column_indexes = range(board.width)
    for column_index in column_indexes:
        # the more this line solved
        # priority = 1 - board.column_solution_rate(column_index)

        # the closer to edge
        # priority = 1 - abs(2.0 * column_index / board.width - 1)

        # the more 'dense' this line
        # priority = 1 - board.densities[True][column_index]

        new_job = (True, column_index)

        priority = 0
        if has_blots:
            # the more attempts the less priority
            priority = board.attempts_to_try(*new_job)

        _add_job(new_job, priority)

    total_cells_solved = 0

    for (is_column, index), priority in line_jobs.sorted_iter():
        # LOG.info('Solving %s %s with priority %s', index,
        #          'column' if is_column else 'row', priority)

        new_jobs = solve_row(board, index, is_column, method)

        total_cells_solved += len(new_jobs)
        for new_job in new_jobs:
            new_priority = priority - 1
            if board.has_blots:
                # the more attempts the less priority
                new_priority = board.attempts_to_try(*new_job)

            # lower priority = more priority
            _add_job(new_job, new_priority)

        lines_solved += 1

    # all the following actions applied only to verified solving
    if not contradiction_mode:
        board.solution_round_completed()

        # rate = board.solution_rate
        # if rate != 1:
        #     LOG.warning('The nonogram is not solved full (%r). The rate is %.4f',
        #                 method, rate)
        logger.info('Full solution: %.6f sec', time.time() - start)
        logger.info('Lines solved: %i', lines_solved)

    return total_cells_solved, all_jobs
